#pragma once
#include "common.h"
#include "datatype_generics.h"
#include "Rescaler.h"
#include "MultiScale2.h"

/** A utility class to embed one gridpacket into another according to the worldspace location.

Uses the rescalar classes, and supports a variety of placement options. This could be used by the File Input device, or other similar types of devices.

*/
namespace WorldMachine {
	namespace Rescale {
		// These are internal helper functions; don't mind them.
		namespace internal {
			template < class T > 
			typename T::value_type kernalAdapter(T &src, CoordF gridloc, RescaleInfo const&rescaletype, FractalRescaleInfo &fr) {
				fr.setGridLocation(gridloc);
				Coord  srcloc(int(gridloc.x), int(gridloc.y));
				CoordF uv;
				uv.x = gridloc.x - srcloc.x;
				uv.y = gridloc.y - srcloc.y;


				T::value_type data[4][4];
				for (int j = 0; j < 4; ++j) {
					for (int i = 0; i < 4; ++i) {
						data[i][j] = src.valueSafe(Coord(srcloc.x+i-1, srcloc.y+j-1));
					}
				}

				return WorldMachine::Rescale::Resample_Kernel(data, uv, rescaletype, fr);
			}

			template < class T > 
			typename T::value_type kernalAdapterTiled(T &src, CoordF gridloc, RescaleInfo const&rescaletype, FractalRescaleInfo &fr) {
				fr.setGridLocation(gridloc);
				assert(gridloc.x >= 0.0f);
				assert(gridloc.y >= 0.0f);

				Coord  srcloc(int(gridloc.x), int(gridloc.y));
				CoordF uv;
				uv.x = gridloc.x - srcloc.x;
				uv.y = gridloc.y - srcloc.y;

				Coord srcsize = src.getsize();

				T::value_type data[4][4];
				for (int j = 0; j < 4; ++j) {
					for (int i = 0; i < 4; ++i) {
						Coord innercoord = Coord(srcloc.x+i-1, srcloc.y+j-1);
						if (innercoord.x < 0)
							innercoord.x += srcsize.x;
						else if (innercoord.x >= srcsize.x)
							innercoord.x -= srcsize.x;
						if (innercoord.y < 0)
							innercoord.y += srcsize.y;
						else if (innercoord.y >= srcsize.y)
							innercoord.y -= srcsize.y;
						data[i][j] = src.value(innercoord);
					}
				}

				return WorldMachine::Rescale::Resample_Kernel(data, uv, rescaletype, fr);
			}

			template < class T > 
			typename T::value_type kernalAdapterMirrored(T &src, CoordF gridloc, RescaleInfo const&rescaletype, FractalRescaleInfo &fr) {
				fr.setGridLocation(gridloc);
				assert(gridloc.x >= 0.0f);
				assert(gridloc.y >= 0.0f);

				Coord  srcloc(int(gridloc.x), int(gridloc.y));
				CoordF uv;
				uv.x = gridloc.x - srcloc.x;
				uv.y = gridloc.y - srcloc.y;

				Coord srcsize = src.getsize();
				Coord mirrorsize = srcsize * 2;

				T::value_type data[4][4];
				for (int j = 0; j < 4; ++j) {
					for (int i = 0; i < 4; ++i) {
						Coord innercoord = Coord(srcloc.x+i-1, srcloc.y+j-1);
						if (innercoord.x < 0)
							innercoord.x += mirrorsize.x;
						else if (innercoord.x >= mirrorsize.x)
							innercoord.x -= mirrorsize.x;
						if (innercoord.y < 0)
							innercoord.y += mirrorsize.y;
						else if (innercoord.y >= mirrorsize.y)
							innercoord.y -= mirrorsize.y;

						if (innercoord.x >= srcsize.x)
							innercoord.x = mirrorsize.x - innercoord.x - 1;
						if (innercoord.y >= srcsize.y)
							innercoord.y = mirrorsize.y - innercoord.y - 1;

						data[i][j] = src.value(innercoord);
					}
				}

				return WorldMachine::Rescale::Resample_Kernel(data, uv, rescaletype, fr);
			}
			CoordF wrapWorldSpace(CoordF &pos, CoordF &wrapsize);


		} // end internal

		enum WRAP_BEHAVIOR { WRAP_NORMAL, WRAP_REPEAT_EDGES, WRAP_TILED, WRAP_MIRRORED };
		/**
		 placePacket allows you to insert a localspace grid packet into a specified region of worldspace. The localspace
		 packet's proper placement is determined by the srcplacement parameter specifying where it should lie in worldspace.

		 You can use a WRAP_BEHAVIOR parameter to specify how to handle areas outside of the placement

		*/
		/*
		template < class PKT >
		void placePacket(boost::intrusive_ptr<typename PKT>const &result, boost::intrusive_ptr<typename PKT>const &src, Rect4F const&srcplacement, RescaleInfo const&rescaletype, CoordF placement_vscale = CoordF(0.0f, 1.0f), WRAP_BEHAVIOR wrap_type = WRAP_NORMAL, bool shared_edges = false) {
			placePacket(*result, *src, srcplacement, rescaletype, placement_vscale, wrap_type, shared_edges);
		}*/

		template < class PKT >
		void placePacket(PKT &result, PKT const &src, Rect4F const&srcplacement, RescaleInfo const&rescaletype, CoordF placement_vscale = CoordF(0.0f, 1.0f), WRAP_BEHAVIOR wrap_type = WRAP_NORMAL, bool shared_edges = false) {
			typedef typename PKT::value_type ValType;

			// We must create the fractal parameters in case the user needs chose that type
			PacketContext srcc = src.getContext();
			srcc.setRenderQuad(srcplacement);
			srcc.setSpatialType(PacketContext::WORLDSPACE);
			FractalRescaleInfo fr(srcc, result.getContext());

			// The following are used only by the tiled and mirrored modes
			CoordF srcres;
			if (shared_edges)
				srcres = i2f(src.getsize() - Coord(1,1));
			else
				srcres = i2f(src.getsize());

			CoordF regionsize = srcplacement.size();
			CoordF oosize = CoordF(1.0f, 1.0f) / regionsize;
			CoordF srcoffset(srcplacement.x0, srcplacement.y0);
			//

			switch (wrap_type) {
			default:
			case WRAP_NORMAL:
				result.visitAll([&](typename PKT::GridLocation &location) {
					CoordF gridloc = srcc.transformFromWorld(location.worldpos);
					if (srcc.testGridCoord(gridloc)) {
						location.dataref = vlerp(ValType(placement_vscale.x), ValType(placement_vscale.y), internal::kernalAdapter(src, gridloc, rescaletype, fr) );
					}
					else
						location.dataref = WorldMachine::datatype::zero<ValType>();
				});
				break;
			case WRAP_REPEAT_EDGES:
				result.visitAll([&](typename PKT::GridLocation &location) {
					CoordF gridloc = srcc.transformFromWorld(location.worldpos);
					location.dataref = vlerp(ValType(placement_vscale.x), ValType(placement_vscale.y), internal::kernalAdapter(src, gridloc, rescaletype, fr) );
				});
				break;
			case WRAP_TILED:
				result.visitAll([&](typename PKT::GridLocation &location) {
					CoordF wrapped = internal::wrapWorldSpace(location.worldpos - srcoffset, regionsize);
					CoordF gridloc = wrapped * oosize * srcres;

					location.dataref = vlerp(ValType(placement_vscale.x), ValType(placement_vscale.y), internal::kernalAdapterTiled(src, gridloc, rescaletype, fr) );
				});
				break;
			case WRAP_MIRRORED:
				result.visitAll([&](typename PKT::GridLocation &location) {
					CoordF wrapped = internal::wrapWorldSpace(location.worldpos - srcoffset, regionsize * 2.0f);
					CoordF gridloc = wrapped * oosize * srcres;

					location.dataref = vlerp(ValType(placement_vscale.x), ValType(placement_vscale.y), internal::kernalAdapterMirrored(src, gridloc, rescaletype, fr) );
				});
				break;
			}
		}

		template < class PKT >
		void placePacket(PKT &result, MultiScalePyramid<typename PKT> const &src, Rect4F const&srcplacement, RescaleInfo const&rescaletype, CoordF placement_vscale = CoordF(0.0f, 1.0f), WRAP_BEHAVIOR wrap_type = WRAP_NORMAL, bool shared_edges = false) {
			typedef typename PKT::value_type ValType;

			int highlevel = 0, lowlevel = 0;
			float blendfactor = 0.0f;

			CoordF regionsize = srcplacement.size();
			CoordF oosize = CoordF(1.0f, 1.0f) / regionsize;
			CoordF srcoffset(srcplacement.x0, srcplacement.y0);

			if (src.queryMipmapDetails(result.getDScale(), srcplacement, highlevel, lowlevel, blendfactor) ) { // use mipmap
				boost::intrusive_ptr<typename PKT> const& mipmapHigh = src.retrieveReducedMap(highlevel);
				boost::intrusive_ptr<typename PKT> const& mipmapLow = src.retrieveReducedMap(lowlevel);

				PacketContext highc = mipmapHigh->getContext(), lowc = mipmapLow->getContext();
				highc.setRenderQuad(srcplacement);
				highc.setSpatialType(PacketContext::WORLDSPACE);
				FractalRescaleInfo high_cfr(highc, result.getContext());
				lowc.setRenderQuad(srcplacement);
				lowc.setSpatialType(PacketContext::WORLDSPACE);
				FractalRescaleInfo low_cfr(lowc, result.getContext());


				CoordF high_srcres, low_srcres;
				if (shared_edges) {
					high_srcres = i2f(mipmapHigh->getsize() - Coord(1,1));
					low_srcres = i2f(mipmapLow->getsize() - Coord(1,1));
				}
				else {
					high_srcres = i2f(mipmapHigh->getsize());
					low_srcres = i2f(mipmapLow->getsize());
				}

				switch (wrap_type) {
				default:
				case WRAP_NORMAL:
					result.visitAll([&](typename PKT::GridLocation &location) {
						CoordF gridlow = lowc.transformFromWorld(location.worldpos);
						CoordF gridhigh = highc.transformFromWorld(location.worldpos);
						if (highc.testGridCoord(gridhigh)) {
							assert(lowc.testGridCoord(gridlow));
							ValType a = internal::kernalAdapter(*mipmapHigh, gridhigh, rescaletype, high_cfr);
							ValType b = internal::kernalAdapter(*mipmapLow, gridlow, rescaletype, low_cfr);
							location.dataref = vlerp(ValType(placement_vscale.x), ValType(placement_vscale.y), lerp(a, b, blendfactor) );
						}
						else
							location.dataref = WorldMachine::datatype::zero<ValType>();
					});
					break;
				case WRAP_REPEAT_EDGES:
					result.visitAll([&](typename PKT::GridLocation &location) {
						CoordF gridlow = lowc.transformFromWorld(location.worldpos);
						CoordF gridhigh = highc.transformFromWorld(location.worldpos);
						ValType a = internal::kernalAdapter(*mipmapHigh, gridhigh, rescaletype, high_cfr);
						ValType b = internal::kernalAdapter(*mipmapLow, gridlow, rescaletype, low_cfr);
						location.dataref = vlerp(ValType(placement_vscale.x), ValType(placement_vscale.y), lerp(a, b, blendfactor) );
					});
					break;
				case WRAP_TILED:
					result.visitAll([&](typename PKT::GridLocation &location) {
						CoordF wrapped = internal::wrapWorldSpace(location.worldpos - srcoffset, regionsize);
						CoordF gridlow = wrapped * oosize * low_srcres;
						CoordF gridhigh = wrapped * oosize * high_srcres;

						ValType a = internal::kernalAdapterTiled(*mipmapHigh, gridhigh, rescaletype, high_cfr);
						ValType b = internal::kernalAdapterTiled(*mipmapLow, gridlow, rescaletype, low_cfr);
						location.dataref = vlerp(ValType(placement_vscale.x), ValType(placement_vscale.y), lerp(a, b, blendfactor) );
					});
					break;
				case WRAP_MIRRORED:
					result.visitAll([&](typename PKT::GridLocation &location) {
						CoordF wrapped = internal::wrapWorldSpace(location.worldpos - srcoffset, regionsize * 2.0f);
						CoordF gridlow = wrapped * oosize * low_srcres;
						CoordF gridhigh = wrapped * oosize * high_srcres;

						ValType a = internal::kernalAdapterMirrored(*mipmapHigh, gridhigh, rescaletype, high_cfr);
						ValType b = internal::kernalAdapterMirrored(*mipmapLow, gridlow, rescaletype, low_cfr);
						location.dataref = vlerp(ValType(placement_vscale.x), ValType(placement_vscale.y), lerp(a, b, blendfactor) );
					});
					break;
				}
			}
			else {	// if at highest mipmap level already, or magnifying, then don't trilinear filter
				boost::intrusive_ptr<typename PKT> map = src.retrieveReducedMap(highlevel);
				placePacket<PKT>(result, *map, srcplacement, rescaletype, placement_vscale, wrap_type, shared_edges);
			}
		}
	}
};