#pragma once

#include <functional>
#include "WMException.h"

namespace WorldMachine {

/**
	C++11 style visitors (iterators) to quickly perform some function across all pixels of a grid data packet.

	YOu can in fact iterate over multiple data packets at the same time, for example if you have several inputs and an output packet.

	Note that the PacketVisitor requires that all packets be of the same type. If you are iterating over non-homogenous datatypes,
	then capture your additional packets in your function object and index into them using the provided PacketVisitor::location.index value.

	How to use:

	1) Instantiate your PacketVisitor by calling the constructor with your datatypes that you want ready access to. The visitor is parameterized on the datatype in question.
	2) call the visitAll() or visitRegion() member functions to visit all locations. Provide a function object (or lambda of course) that accepts a PacketVisitor<>::Location object.

	so for example:
	@code

	PacketVisitor<HField> visit(myHF);
	// optional: provide a progress callback if you want the processing to visitor to report iteration progress and be interruptible.
  	visit.setStandardCallback(YourDevice, YourBuildContext);

	visit.visitAll([] ( PacketVisitor<HField>::Location const& loc ) { 
	  // loc contains references, pointers and location information for us. The most convenient form is like so:
	  loc[0] = somevalue; // the array overload allows you to specify which underlying packet you wish to operate on. Since we have only 1 packet in the current visitor, we operate
						  // on entry 0, at the current location.
						  // 
	  loc[1] = somethingelse; // ERROR: This will throw a WMException as we only have a single heightfield in the visitor. 
	  *loc.itr[0] = somevalue; // OK
	  myHF.value[loc.loc]; // OK, use the grid location provided by the iterator
	});

	@endcode

	Notes:
	 Ideally, this would use variadic templates so that the packetvisitor could contain multiple different packet types, rather than be specialized on only one type.
	 However, the current MSVC2010 compiler does not support variadic templates. Perhaps in the next iteration.. 

 */
template < class PacketType >
class PacketVisitor {
public:
	typedef typename PacketType* PacketPtr;
	typedef std::function<bool (int, int)> ProgressFunc;
	PacketVisitor(PacketType &first) { firstpacket(first); };
	PacketVisitor(PacketType &first, PacketType &second) {firstpacket(first); add(second); };
	PacketVisitor(PacketType &first, PacketType &second, PacketType &third) {firstpacket(first); add(second); add(third); };
	PacketVisitor(PacketType &first, PacketType &second, PacketType &third, PacketType &fourth) {firstpacket(first); add(second); add(third); add(fourth); };
	// If you need to create with more than 4 packets, call add() yourself to insert the extras.


	/// A default callback class that can be supplied if you have nothing more complicated to do than simply do a single iteration through your height field.
	class StandardDeviceCallback {
	public:
		StandardDeviceCallback(Device *_dev, BuildContext &_context) : context(_context), dev(_dev) {};
		bool operator()(int val, int max) { return context.ReportDeviceProgress(dev, val, max); };
	private:
		Device *dev;
		BuildContext &context;
	};

	/// Location is passed to your visitor function and provides iterators to each of the established packets in the set, as well as an index, heightfield, and worldspace coordinates.
	/// Note that Location is not copyable -- you may only receive it by reference!
	struct Location {
		Location() {}
		typename PacketType::value_type& operator[](size_t index) const {
			if (index >= itr.size())
				throw WMException("Iterator index value too large!");
			return *(itr[index]);
		}

		/// The current world space coordinate
		CoordF wloc;
		/// Your current grid coordinate
		Coord loc;
		/// The index into the grid data
		size_t index;
		/// vector of iterators. These are updated by the visit function to always point to the current grid cell for all of your maps.
		std::vector<typename PacketType::iterator> itr;
	protected:
		// cannot copy
//		Location(Location &other);
//		Location &operator=(Location &other);
	};

	/// Attach a custom callback function that will be called periodically during the build process
	void setProgressCallback(ProgressFunc const& progress) {progress_callback = progress; };

	/// Creates and attaches simple callback function for you based upon the current build context and device.
	void setStandardCallback( Device* dev, BuildContext & context ) 
	{
		setProgressCallback(PacketVisitor<HField>::StandardDeviceCallback(dev, context));
	}


	/// Add a PacketType to the set. Validates that he packet being added is the correct size
	size_t add(PacketType &packet) { if (areacontext.getRes() != packet.getsize()) { throw WMException("Size of packets don't match in PacketVisitor"); }; packets.push_back(&packet); return packets.size()-1; };

	/// VisitFunc will be supplied with a PacketVisitor<PacketType>::Location object
	template < typename VisitFunc >
	bool visitAll( VisitFunc visitor ) {
		RectF area = areacontext.getRenderExtent();
		CoordF scaling(area.width / (areacontext.w()-1), area.height / (areacontext.h()-1));

		size_t packetsize = packets.size();

		Location loc;
		loc.wloc = CoordF(area.x, area.y);
		loc.loc = Coord(0,0);
		loc.index = 0;
		loc.itr.resize(packetsize);
		for (size_t i = 0; i < packetsize; ++i) {
			loc.itr[i] = packets[i]->GetDataPtr();
		}

		int w = areacontext.w();
		int h = areacontext.h();

		// Do the traversal with manual pointer arithmetic while updating the world and HF-space coordinates appropriately
		for (int y = 0; y < h; y++) {
			loc.wloc.x  = area.x;
			loc.loc.x = 0;
			for (int x = 0; x < w; x++) {
				visitor(loc); // invoke

				// update all locs and ptrs
				loc.wloc.x += scaling.x;
				++loc.loc.x;
				++loc.index;
				for (size_t i = 0; i < packetsize; i++) {
					++loc.itr[i];
				}
			}
			++loc.loc.y;
			loc.wloc.y += scaling.y;
			// call the progress functor every scanline, and break if it indicates to abort.
			if (progress_callback && !progress_callback(y, h))
				return false;
		}
		return true;
	}
protected:
	void firstpacket(PacketType &first) {
		assert(packets.empty());
		areacontext = first.getContext();
		if (areacontext.resarea() == 0)
			throw WMException("Instantiated with empty packet"); 
		add(first);
	}

	std::vector<PacketPtr> packets;
	PacketContext areacontext;
	ProgressFunc progress_callback;
};

}