#pragma once
#include "common.h"
#include <vector>

/**	@file Parameter.h Parameter-related classes and defines
*/

class Device;

const int PARAM_NAME_SIZE = 32;
const int GROUP_NAME_SIZE = 32;

/// @name Widget Types
/// This provide hints to the application layer asto the best way to display the parameter.
/// @relates Parameter
//@{
const int WIDGET_SLIDER = 0;					///< provide a slider (default)
const int WIDGET_LOGSLIDER = 1;					///< logarithmic scaling on the slider
const int WIDGET_EDITONLY = 2;					///< provide only an edit box (no slider
const int WIDGET_SLIDER_FLAG_METERS = 512;		///< display slider values in WM units or kilometers as per current preference
const int WIDGET_SLIDER_FLAG_HEIGHT = 256;		///< Specify as a WM vertical height unit (ie WM 0..1 or m/km as per pref)
//@}

/// The possible value types for a Parameter
/// @relatesalso Parameter
enum V_TYPE { V_INT, V_FLOAT, V_BOOL, V_PTR, V_ENUM, V_STR, V_COLOR };

/** for type V_PTR, the parameter is an action. The pointer stored is expected to be the
	address of a standard call function that accepts one void* parameter. The framework will pass
	in the Device* pointer of the current device to the supplied function. Your function can
	then call a member function of the device, or do whatever other action is desired.
*/
typedef void (*VPtrType)(Device *owner);

/// The value of a parameter. Access the given type according to the V_TYPE of the parameter
/// @relates Parameter
union VAL {
	VAL() { };
	VAL(int _i) { i = _i; };
	VAL(float _f) { f = _f; };
	VAL(bool _b) { b = _b; };

	int i;
	float f;
	bool b;
	VPtrType vptr;		// in 64bit system, will make sizeof(VAL) == 64bits
	dword color;
};

class EnumType {
public:
	char name[PARAM_NAME_SIZE];
};

/// @name Parameter flags
/// 
/// @relates Parameter
//@{
const int PARM_FLAG_HIDDEN		= 1;	// Do not assign a port or show this param
const int PARM_FLAG_DEPRECATED	= 2;	// Do not assign a port or allow user to change this param
const int PARM_FLAG_CLIP		= 4;	// Instead of scaling the override value to min..max, clip it
const int PARM_FLAG_READONLY	= 8;	// Don't allow the user to modify this parameter. (Usually used for feedback to the user)
const int PARM_FLAG_CENTER		= 16;	// Center the parameter instead of left-justifying it
const int PARM_FLAG_NODELETE	= 32;	// Dont allow the user to delete this parameter. This is UNUSED at the moment!
const int PARM_FLAG_NOPORT		= 64;	// Do not assign a port to this param
const int PARM_FLAG_DISABLED	= 128;	// Disabled by Dependency
//@}

/** Parameters provide an easy way to store and manage basic Device configuration data.
	Parameters can only be one of:
	scalars, integers, colors, enums, strings, function pointers

	What's nice is that each device by default contains a set of Parameters, allowing the framework to handle parameter
	serialization.

	@ingroup devices
*/
class Parameter {
public:
/// @name Creation
/// Each contructor creates a different type of Parameter
//@{ 
// generic constructors
	Parameter();
	Parameter(const char *_name);
// specific constructors
	Parameter(const char *_name, int _value, int _min, int _max, int _wtype = WIDGET_SLIDER);
	Parameter(const char *_name, float _value, float _min, float _max, int _wtype = WIDGET_SLIDER);
	Parameter(const char *_name, bool _value);
	Parameter(const char *_name, VPtrType _ptr);	// action (button) parameter
												// see notes above regarding call procedure
	Parameter(const Parameter&);
	~Parameter();
	Parameter& operator= (const Parameter& other);
//@}

	/// Load parameter from disk.
	bool Load(std::istream &in);
	/// Save parameter to disk
	bool Save(std::ostream &out);

	bool LoadValue(std::istream &in);	///< Reads only the current value. Assumes everything except the "value" parameter is already setup
	bool SaveValue(std::ostream &out);	///< Writes only the current Value, nothing more.

	void CopyStripped(const Parameter &other);	///< Copy only the vitals; dont copy the strings. Faster.

	void ClampValue();	///< Enforces that the current value is between min..max

	/// @name Data I/O Primitives
	/// Use the one appropriate to the type of the Parameter.
	//@{
	int getInt()		{ assert(type == V_INT || type == V_ENUM); return value.i; };
	float getFloat()	{ assert(type == V_FLOAT); return value.f; };
	bool getBool()		{ assert(type == V_BOOL); return value.b; };
	dword getColor()	{ assert(type == V_COLOR); return value.color; };
	const char*getString() { assert(type == V_STR); return getHelpString(); };

	void setInt(int a) { value.i = a; };
	void setFloat(float f) { value.f = f; };
	void setBool(bool b) { value.b = b; };
	void setColor(dword c)	{ value.color = c; };
	void setString(const char*str) { assert(type == V_STR); setHelpString(str); };
	//@}

/// @name Enum Functions
/// A set of contiguous integer values starting at 0, each of which has an associated string.
/// Useful for named parameter options. 
//@{
	void SetupEnum(int entries = -1); ///< If you don't provide the 'entries' parameter, you must first set the parameter's maximum range value. 
	int  numEnum();
	bool setEnumString(const int val, const char *str);
	char *getEnumString(const int val);
//@}

	void setHelpString(const char *str);	///< Set the help string for this parameter
	const char *getHelpString();			///< Retrieve help string for this parameter if it exists

// Get the flag bits
	int  getFlags()			 { return flags; };
	void setFlag(int _flags) { flags = _flags; };
// Check some specific flag bits
	bool isVisible()			{ return !(flags & (PARM_FLAG_HIDDEN | PARM_FLAG_NOPORT | PARM_FLAG_DEPRECATED)); };
	bool isEnabled()			{ return !(flags & PARM_FLAG_DISABLED); };
	// Set specific flags
	void setEnabled(bool en)	{ if (!en) flags |= PARM_FLAG_DISABLED; else flags &= ~PARM_FLAG_DISABLED; };

	VAL value;					///< the current value
	VAL min, max;				///< the range of acceptable values
	V_TYPE type;				///< the type of VTYPE
	int widget_type;			///< how should it be shown & adjusted?
	char name[PARAM_NAME_SIZE];	///< name of this parameter
private:
	int flags;
	std::vector<EnumType> names;// storage of enumations
	char *help_str;				// pointer to help text for this parameter, if any
};

const unsigned char PARAM_SYNCH = 0x0ce;

/** A simple class that provides a way to group multiple, related parameters together.
 TODO: This class is currently under construction and should not be used yet.
*/
class ParamGroup {
public:
	void setName(const char*_name);///M Set the groupname

	char name[GROUP_NAME_SIZE];	///< group name
	int group_start;			///< which param starts this group
	int group_end;				///< which param ends this group
};

/// Dependency type. VALUE_CHANGE is for the dependency function and is called whenever the dependent value changes.
/// The order of the following is dependsOn, dependsValue. For example:
///  VALUE_GREATER means that dependsOn > dependsValue
enum DEPENDS_TYPE { VALUE_EQUAL, VALUE_CHANGE, VALUE_NOTEQUAL, VALUE_GREATER, VALUE_LESS, VALUE_GREATEREQUAL, VALUE_LESSEQUAL };

struct DependsPair;

typedef void (*DEPENDS_FUNC)(Device *dev, DependsPair &depair);

/// Parameter Dependency datastructure. Used internally by ParameterSet
struct DependsPair {
public:
	int dependsOwner;		// the index of the parameter that will be changed (target)
	int dependsOn;			// the index of the parameter that will trigger this dependency when it changes
	VAL value;				// the value that should trigger this dependence. The type of VAL is the type of dependsOn
	DEPENDS_TYPE type;		// the type of dependence
	DEPENDS_FUNC func;		// optional function to call upon dependency update
};

/** A collection of Parameters

	Used by Devices to contain all of their configuration details in an automatic way. All parameters controlled by the Parameter and
	ParameterSet classes are automatically copied, saved, and loaded, and have generic dialog support by the World Machine framework.

	Thus, unless your device data cannot be stored easily in parameters, it is recommended to use these classes!
	@ingroup devices
*/
class ParameterSet
{
public:
	ParameterSet(void);
	~ParameterSet(void);

	bool Load(std::istream &in);
	bool Save(std::ostream &out);

	void AddParam(Parameter const&p);
	bool DelParam(int index);
	bool UpdateParamValue(const Parameter &p);				/// Update parameter, searching by name
	bool UpdateParamValue(int index, const Parameter &p);	/// Update value of existing parameter
	bool ReplaceParam(int index, const Parameter &newparam);/// Replace with new parameter
	void ResetParams();									///< Removes all parameters

	Parameter *GetParam(int index);
	Parameter *GetParam(const char *name);
	int		   GetParamIndex(const char *name);

/// @name Accessor Functions
//@{
	int		getInt(int i); 
	float	getFloat(int i);
	bool	getBool(int i);
	int		getInt(const char *name);
	float	getFloat(const char *name);
	bool	getBool(const char *name);
//@}

/// Number of parameters
	int size() const { return (int) params.size(); };
/// Assign this set of parameters to its owner. You will not typically need to do this.
	void setOwner(Device *_spr) { owner = _spr; };
	Device *getOwner() const { return owner; };

/// Add an already-defined ParamGtroup to the set
	void AddGroup(ParamGroup &group) { groups.push_back(group); };

/// Execute dependency checking, changing the state of any parameters that depend on the current state of another.
/// Specify the changed parameter or leave blank to rescan all parameters
	bool ProcessDependencies(const int changedindex = -1);
/// Add a Parameter Dependency
	bool AddDependence(const char *dependsOwner, const char *dependsOn, DEPENDS_TYPE type, VAL dependsval, DEPENDS_FUNC func = NULL);

	bool allow_noncurrent_parms;	///< Control whether on parameter load we allow parameters different from those already defined to be created.
private:
	Device *owner;
	std::vector<Parameter> params;
	std::vector<DependsPair> dependence;
	std::vector<ParamGroup> groups;
};