#pragma once

#include <iostream>
#include <sstream>
#include <stdint.h>
#include <iosfwd>
#include <map>
#include "WMAllocator.h"

namespace WorldMachine {
/**

A helper class to more easily produce or consume the new (as of WM 2.3.7) stream-oriented version tagged files.

The way this is intended to work:

Each exportable class (has Load/Save() ) will define string tags that delineate storage for a particular group of features. As WM evolves, new groups of features are 
added within a tag group; older versions of WM can skip reading any sections they don't understand (although this obviously may violate intended functioning of the world,
and should be warned about!).

This can be thought of as a stream-friendly tagged file format a bit similar to a very simplified TIFF, but explicitly not requiring the reader or writer to seek around in the file.

The main reason to move to this system instead of a rigid version-numbered system is to allow better interoperability of files between the dev and release channels;
people using the dev channel shouldn't be disallowed to share macros, etc with release channel people. Older versions will simply ignore what they can't grok (if anything).

Implementation details:

Any WM file now consists of:

1) Magic Header section
2) One or more tagged section blocks

Each tag header consists of:

1) A 2byte magic marker specifying the start of a tag. This helps the reader know they will be parsing valid data.
2) A 1byte tag length value. This specifies the length of the tagname. Max value of 255.
3) An 8byte size field. 64bit size fields were chosen over 32bit as some use cases require the ability to store very large >4GB blocks (ex: storage of data files)
4) An n-byte TAGNAME. These need not be unique; however, you must define whatever the semantic are of having multiple tags of the same type at the same heirarchy level.

The tag data block follows this header. To find the next tag block, advance size number of bytes from the first byte of the datablock.

There is no requirement that the data within a tag section be anything but implementation-specified; however sometimes it is useful to write entities using the same format,
easing the addition of new features, etc. If you do this, it's recommended to not write raw values to the tag stream unless it is a leaf entity.
*/

namespace io {
	const int16_t TAGSECTION_MAGIC_TAG = 0x04d57; // 'WM'
	typedef std::basic_stringstream<char, std::char_traits<char>, wmcore_allocator<char> > wmstringstream;

/** A TagEntity is a C++ representation of a tag section in a file. It consists of a tag name and a stream buffer that can receive or provide data.

The TagEntity class can be used to:
 * form section entries in a heirarchical way;
 * preserve unknown tags to be saved again with the file even when we can't use them.
 * hold any arbitrary binary blob data for later processing.

 Note that since the TagEntity copies its contents in memory, you should refrain from storing large binary blobs (more than 100kb, say) with this class to
 avoid consuming unneeded memory and CPU -- instead use the low level writeTagHeader directly to an output stream.
*/
class TagEntity {
public:
	TagEntity();
	/// Construct a tag entry for writing
	explicit TagEntity( const char *tagname);
	/// Copy a tag entry
	explicit TagEntity( TagEntity const&other);

	/// Read a tagentity from the input stream. The entirety of the entity will be read; the stream pointer will be on the next byte after this entity.
	explicit TagEntity( std::istream &instream);

	/// Read a tagentity from the input stream. The tagname must be the provided matchingtag; if not, a ComponentIOException will be thrown.
	explicit TagEntity( std::istream &instream, const char *matchingtag);

	/// Prepare a new tagged entity for IO, using the given tag. Previous contents will be deleted.
	void Create(const char *tagname);

	/// Add a child TagEntity to this one. The contents of the child entity will be copied into this entity's datablock.
	void addChild(TagEntity const&entity);

	/// Convenience function to create and add a child TagEntity consisting of the provided raw binary datablock. It will be tagged with the provided name and
	/// added as a child to this tag.
	template <typename T>
	void addChildValue(const char *tagname, T const&data, size_t len = sizeof(T));

	/// Retrieve tagname
	WMString getTag() { return tag; };

	/// Write the current contents of this entity to an output stream
	void writeTo(std::ostream &out) const;

	/// Retrieve an iostream to read/write into this entity.
	operator std::iostream &() { return internalstream; };

	// Retrieve the bound stringstream object for direct string access
	io::wmstringstream &stream() { return internalstream; };
protected:
	void setTag(WMString const& newtag) { tag = newtag; };
	void clearInternals();
private:
	WMString tag;
	io::wmstringstream internalstream;
};

/** Convenience class to parse children of a tag entity.

	TagParser parser;
	parser.parseChildEntities(parent);

	bool entity_exists = parser.hasValue("Abcd");

	TagEntity e = parser.getChildValue("Abcd");

*/
class TagParser {
public:
	TagParser();	///< default construct for 2-part construction, when you need to add tag handlers before parsing the data.
	/// Part 2 of a 2-step construction. Read all child tag entities. Note that any tags that have a custom handler defined will
	/// NOT be included in this set -- this is so they can be handled individually.
	void parseChildEntities(TagEntity &te);

	/// Convenient one-step parser factory. Creates the parser and executes the parse. Easiest method, useful when you have no special handlers to add.
	static TagParser createParser(TagEntity &parent);
	/// Create an internal tag entity from the stream and parse it. The tagname must be the provided matchingtag; if not, a ComponentIOException will be thrown.
	static TagParser createParser(std::istream &stream, const char*tagname);

	/// Add a custom handler for a given tag. This is most useful for handling tags that denote large blob data that shouldn't be copied.
	void addTagHandler(const char*tag, std::function<void(TagEntity&)> func);

	/// Check for existence of a child
	bool hasValue(const char*tagname);

	/// Retrieve a pointer to the entity of a child value, or nullptr if it doesn't exist.
	/// You can only retrieve values after parsing!
	TagEntity* getValue(const char*tagname);

	/// Read the results into the child value, if it exists.
	/// Returns true if results were read.
	template < class T > 
	bool readValue(const char*tagname, T &result, size_t count = sizeof(T)) {
		// Supplying a pointer is always an error and will be flagged as such.
		static_assert(!(std::is_pointer<T>::value), "You must supply a reference to the value itself, not a pointer to the value!");
		TagEntity *v = getValue(tagname);
		if (v) {
			v->stream().read(reinterpret_cast<char*>(&result), count);
		}
		return !!v;
	};

private:
	typedef std::map<WMString, TagEntity> TagSet;
	typedef std::pair<WMString, TagEntity> TagSetPair;
	typedef std::pair<const char*, std::function<void(TagEntity&)> > handlertype;
	std::vector<handlertype> handlers;
	TagSet values;
};

// Miscellaneous helper functions

/// Write a tag header to the output stream. This is the lowest-level of tagged IO other than handling everything yourself at the byte level.
/// Must supply a WMString and a tagdata size value. Your binary blob should be then written to the output stream following this header.
void writeTagHeader(std::ostream &out, WMString const&name, const size_t tagsize);

/// Check and see if the next several bytes in the stream define a valid tag header
bool probeTagHead(std::istream &in);


// *** Implementations ***


/// Convenience function to write out non-char data as a tagged entity.
/// This is for use only with plain old data types, not types with embedded non-POD members
/// You should supply the value type itself to the write function, not a pointer, for example:
/// int a = 6532;	// sizeof(int) == 4 here
/// write(out, a); // write the value of a to stream. 4 bytes will be written.
template <typename T>
void WorldMachine::io::TagEntity::addChildValue( const char *tagname, T const&data, size_t len /*= sizeof(T)*/ )
{
	// note that we allow T to be of array type as taking the address of an array type does NOT decay into a pointer, thus
	// allowing writes to Just Work in this case.
	// Supplying a pointer is always an error and will be flagged as such.
	static_assert(!(std::is_pointer<T>::value), "You must supply a reference to the value itself, not a pointer to the value!");

	TagEntity te(tagname);
	te.stream().write(reinterpret_cast<const char*>(&data), len);
	this->addChild(te);
}


}
}
